@@ -1,4 +1,5 @@
 {- git repository recovery
+import qualified Data.Set as S
  -
  - Copyright 2013 Joey Hess <joey@kitenet.net>
  -
@@ -12,10 +13,11 @@ module Git.Repair (
 	cleanCorruptObjects,
 	retrieveMissingObjects,
 	resetLocalBranches,
-	removeTrackingBranches,
 	checkIndex,
+	checkIndexFast,
 	missingIndex,
 	emptyGoodCommits,
+	isTrackingBranch,
 ) where
 
 import Common
@@ -25,6 +27,7 @@ import Git.Objects
 import Git.Sha
 import Git.Types
 import Git.Fsck
+import Git.Index
 import qualified Git.Config as Config
 import qualified Git.Construct as Construct
 import qualified Git.LsTree as LsTree
@@ -42,17 +45,16 @@ import qualified Data.ByteString.Lazy as L
 import Data.Tuple.Utils
 
 {- Given a set of bad objects found by git fsck, which may not
- - be complete, finds and removes all corrupt objects, and
- - returns a list of missing objects, which need to be
- - found elsewhere to finish recovery.
+ - be complete, finds and removes all corrupt objects,
+ - and returns missing objects.
  -}
-cleanCorruptObjects :: FsckResults -> Repo -> IO (Maybe MissingObjects)
+cleanCorruptObjects :: FsckResults -> Repo -> IO FsckResults
 cleanCorruptObjects fsckresults r = do
 	void $ explodePacks r
 	objs <- listLooseObjectShas r
 	mapM_ (tryIO . allowRead . looseObjectFile r) objs
 	bad <- findMissing objs r
-	void $ removeLoose r $ S.union bad (fromMaybe S.empty fsckresults)
+	void $ removeLoose r $ S.union bad (knownMissing fsckresults)
 	-- Rather than returning the loose objects that were removed, re-run
 	-- fsck. Other missing objects may have been in the packs,
 	-- and this way fsck will find them.
@@ -98,20 +100,17 @@ explodePacks r = do
  - If another clone of the repository exists locally, which might not be a
  - remote of the repo being repaired, its path can be passed as a reference
  - repository.
- 
- - Can also be run with Nothing, if it's not known which objects are
- - missing, just that some are. (Ie, fsck failed badly.)
  -}
-retrieveMissingObjects :: Maybe MissingObjects -> Maybe FilePath -> Repo -> IO (Maybe MissingObjects)
+retrieveMissingObjects :: FsckResults -> Maybe FilePath -> Repo -> IO FsckResults
 retrieveMissingObjects missing referencerepo r
-	| missing == Just S.empty = return $ Just S.empty
+	| not (foundBroken missing) = return missing
 	| otherwise = withTmpDir "tmprepo" $ \tmpdir -> do
 		unlessM (boolSystem "git" [Params "init", File tmpdir]) $
 			error $ "failed to create temp repository in " ++ tmpdir
 		tmpr <- Config.read =<< Construct.fromAbsPath tmpdir
 		stillmissing <- pullremotes tmpr (remotes r) fetchrefstags missing
-		if stillmissing == Just S.empty
-			then return $ Just S.empty
+		if S.null (knownMissing stillmissing)
+			then return stillmissing
 			else pullremotes tmpr (remotes r) fetchallrefs stillmissing
   where
 	pullremotes tmpr [] fetchrefs stillmissing = case referencerepo of
@@ -121,12 +120,12 @@ retrieveMissingObjects missing referencerepo r
 				void $ explodePacks tmpr
 				void $ copyObjects tmpr r
 				case stillmissing of
-					Nothing -> return $ Just S.empty
-					Just s -> Just <$> findMissing (S.toList s) r
+					FsckFailed -> return $ FsckFailed
+					FsckFoundMissing s -> FsckFoundMissing <$> findMissing (S.toList s) r
 			, return stillmissing
 			)
 	pullremotes tmpr (rmt:rmts) fetchrefs ms
-		| ms == Just S.empty = return $ Just S.empty
+		| not (foundBroken ms) = return ms
 		| otherwise = do
 			putStrLn $ "Trying to recover missing objects from remote " ++ repoDescribe rmt ++ "."
 			ifM (fetchfrom (repoLocation rmt) fetchrefs tmpr)
@@ -134,10 +133,10 @@ retrieveMissingObjects missing referencerepo r
 					void $ explodePacks tmpr
 					void $ copyObjects tmpr r
 					case ms of
-						Nothing -> pullremotes tmpr rmts fetchrefs ms
-						Just s -> do
+						FsckFailed -> pullremotes tmpr rmts fetchrefs ms
+						FsckFoundMissing s -> do
 							stillmissing <- findMissing (S.toList s) r
-							pullremotes tmpr rmts fetchrefs (Just stillmissing)
+							pullremotes tmpr rmts fetchrefs (FsckFoundMissing stillmissing)
 				, pullremotes tmpr rmts fetchrefs ms
 				)
 	fetchfrom fetchurl ps = runBool $
@@ -190,15 +189,17 @@ resetLocalBranches missing goodcommits r =
 			, Param (show c)
 			] r
 
+isTrackingBranch :: Ref -> Bool
+isTrackingBranch b = "refs/remotes/" `isPrefixOf` show b
+
 {- To deal with missing objects that cannot be recovered, removes
- - any remote tracking branches that reference them. Returns a list of
- - all removed branches.
+ - any branches (filtered by a predicate) that reference them
+ - Returns a list of all removed branches.
  -}
-removeTrackingBranches :: MissingObjects -> GoodCommits -> Repo -> IO ([Branch], GoodCommits)
-removeTrackingBranches missing goodcommits r =
-	go [] goodcommits =<< filter istrackingbranch <$> getAllRefs r
+removeBadBranches :: (Ref -> Bool) -> MissingObjects -> GoodCommits -> Repo -> IO ([Branch], GoodCommits)
+removeBadBranches removablebranch missing goodcommits r =
+	go [] goodcommits =<< filter removablebranch <$> getAllRefs r
   where
-  	istrackingbranch b = "refs/remotes/" `isPrefixOf` show b
 	go removed gcs [] = return (removed, gcs)
 	go removed gcs (b:bs) = do
 		(ok, gcs') <- verifyCommit missing gcs b r
@@ -230,7 +231,7 @@ explodePackedRefsFile r = do
 		nukeFile f
   where
 	makeref (sha, ref) = do
-		let dest = localGitDir r ++ show ref
+		let dest = localGitDir r </> show ref
 		createDirectoryIfMissing True (parentDir dest)
 		unlessM (doesFileExist dest) $
 			writeFile dest (show sha)
@@ -338,35 +339,42 @@ verifyTree missing treesha r
 {- Checks that the index file only refers to objects that are not missing,
  - and is not itself corrupt. Note that a missing index file is not
  - considered a problem (repo may be new). -}
-checkIndex :: MissingObjects -> Repo -> IO Bool
-checkIndex missing r = do
-	(bad, _good, cleanup) <- partitionIndex missing r
+checkIndex :: Repo -> IO Bool
+checkIndex r = do
+	(bad, _good, cleanup) <- partitionIndex r
 	if null bad
 		then cleanup
 		else do
 			void cleanup
 			return False
 
+{- Does not check every object the index refers to, but only that the index
+ - itself is not corrupt. -}
+checkIndexFast :: Repo -> IO Bool
+checkIndexFast r = do
+	(indexcontents, cleanup) <- LsFiles.stagedDetails [repoPath r] r
+	length indexcontents `seq` cleanup
+
 missingIndex :: Repo -> IO Bool
 missingIndex r = not <$> doesFileExist (localGitDir r </> "index")
 
-partitionIndex :: MissingObjects -> Repo -> IO ([LsFiles.StagedDetails], [LsFiles.StagedDetails], IO Bool)
-partitionIndex missing r = do
+{- Finds missing and ok files staged in the index. -}
+partitionIndex :: Repo -> IO ([LsFiles.StagedDetails], [LsFiles.StagedDetails], IO Bool)
+partitionIndex r = do
 	(indexcontents, cleanup) <- LsFiles.stagedDetails [repoPath r] r
-	let (bad, good) = partition ismissing indexcontents
-	return (bad, good, cleanup)
-  where
-	getblob (_file, Just sha, Just _mode) = Just sha
-	getblob _ = Nothing
-	ismissing = maybe False (`S.member` missing) . getblob
+	l <- forM indexcontents $ \i -> case i of
+		(_file, Just sha, Just _mode) -> (,) <$> isMissing sha r <*> pure i
+		_ -> pure (False, i)
+	let (bad, good) = partition fst l
+	return (map snd bad, map snd good, cleanup)
 
 {- Rewrites the index file, removing from it any files whose blobs are
  - missing. Returns the list of affected files. -}
-rewriteIndex :: MissingObjects -> Repo -> IO [FilePath]
-rewriteIndex missing r
+rewriteIndex :: Repo -> IO [FilePath]
+rewriteIndex r
 	| repoIsLocalBare r = return []
 	| otherwise = do
-		(bad, good, cleanup) <- partitionIndex missing r
+		(bad, good, cleanup) <- partitionIndex r
 		unless (null bad) $ do
 			nukeFile (indexFile r)
 			UpdateIndex.streamUpdateIndex r
@@ -380,9 +388,6 @@ rewriteIndex missing r
 			UpdateIndex.stageFile sha blobtype file r
 	reinject _ = return Nothing
 
-indexFile :: Repo -> FilePath
-indexFile r = localGitDir r </> "index"
-
 newtype GoodCommits = GoodCommits (S.Set Sha)
 
 emptyGoodCommits :: GoodCommits
@@ -431,40 +436,40 @@ preRepair g = do
 	validhead s = "ref: refs/" `isPrefixOf` s || isJust (extractSha s)
 
 {- Put it all together. -}
-runRepair :: Bool -> Repo -> IO (Bool, MissingObjects, [Branch])
-runRepair forced g = do
+runRepair :: (Ref -> Bool) -> Bool -> Repo -> IO (Bool, [Branch])
+runRepair removablebranch forced g = do
 	preRepair g
 	putStrLn "Running git fsck ..."
 	fsckresult <- findBroken False g
 	if foundBroken fsckresult
-		then runRepair' fsckresult forced Nothing g
+		then runRepair' removablebranch fsckresult forced Nothing g
 		else do
 			putStrLn "No problems found."
-			return (True, S.empty, [])
+			return (True, [])
 
-runRepairOf :: FsckResults -> Bool -> Maybe FilePath -> Repo -> IO (Bool, MissingObjects, [Branch])
-runRepairOf fsckresult forced referencerepo g = do
+runRepairOf :: FsckResults -> (Ref -> Bool) -> Bool -> Maybe FilePath -> Repo -> IO (Bool, [Branch])
+runRepairOf fsckresult removablebranch forced referencerepo g = do
 	preRepair g
-	runRepair' fsckresult forced referencerepo g
+	runRepair' removablebranch fsckresult forced referencerepo g
 
-runRepair' :: FsckResults -> Bool -> Maybe FilePath -> Repo -> IO (Bool, MissingObjects, [Branch])
-runRepair' fsckresult forced referencerepo g = do
+runRepair' :: (Ref -> Bool) -> FsckResults -> Bool -> Maybe FilePath -> Repo -> IO (Bool, [Branch])
+runRepair' removablebranch fsckresult forced referencerepo g = do
 	missing <- cleanCorruptObjects fsckresult g
 	stillmissing <- retrieveMissingObjects missing referencerepo g
 	case stillmissing of
-		Just s
+		FsckFoundMissing s
 			| S.null s -> if repoIsLocalBare g
-				then successfulfinish S.empty []
-				else ifM (checkIndex S.empty g)
-					( successfulfinish s []
+				then successfulfinish []
+				else ifM (checkIndex g)
+					( successfulfinish []
 					, do
 						putStrLn "No missing objects found, but the index file is corrupt!"
 						if forced
 							then corruptedindex
-							else needforce S.empty
+							else needforce
 					)
 			| otherwise -> if forced
-				then ifM (checkIndex s g)
+				then ifM (checkIndex g)
 					( continuerepairs s
 					, corruptedindex
 					)
@@ -473,20 +478,22 @@ runRepair' fsckresult forced referencerepo g = do
 						[ show (S.size s)
 						, "missing objects could not be recovered!"
 						]
-					unsuccessfulfinish s
-		Nothing
-			| forced -> ifM (pure (repoIsLocalBare g) <||> checkIndex S.empty g)
+					unsuccessfulfinish
+		FsckFailed
+			| forced -> ifM (pure (repoIsLocalBare g) <||> checkIndex g)
 				( do
-					missing' <- cleanCorruptObjects Nothing g
+					missing' <- cleanCorruptObjects FsckFailed g
 					case missing' of
-						Nothing -> return (False, S.empty, [])
-						Just stillmissing' -> continuerepairs stillmissing'
+						FsckFailed -> return (False, [])
+						FsckFoundMissing stillmissing' ->
+							continuerepairs stillmissing'
 				, corruptedindex
 				)
-			| otherwise -> unsuccessfulfinish S.empty
+			| otherwise -> unsuccessfulfinish
   where
 	continuerepairs stillmissing = do
-		(remotebranches, goodcommits) <- removeTrackingBranches stillmissing emptyGoodCommits g
+		(removedbranches, goodcommits) <- removeBadBranches removablebranch stillmissing emptyGoodCommits g
+		let remotebranches = filter isTrackingBranch removedbranches
 		unless (null remotebranches) $
 			putStrLn $ unwords
 				[ "Removed"
@@ -498,12 +505,12 @@ runRepair' fsckresult forced referencerepo g = do
 			"Reset these local branches to old versions before the missing objects were committed:"
 		displayList (map show deletedbranches)
 			"Deleted these local branches, which could not be recovered due to missing objects:"
-		deindexedfiles <- rewriteIndex stillmissing g
+		deindexedfiles <- rewriteIndex g
 		displayList deindexedfiles
 			"Removed these missing files from the index. You should look at what files are present in your working tree and git add them back to the index when appropriate."
 		let modifiedbranches = resetbranches ++ deletedbranches
 		if null resetbranches && null deletedbranches
-			then successfulfinish stillmissing modifiedbranches
+			then successfulfinish modifiedbranches
 			else do
 				unless (repoIsLocalBare g) $ do
 					mcurr <- Branch.currentUnsafe g
@@ -517,36 +524,36 @@ runRepair' fsckresult forced referencerepo g = do
 								]
 				putStrLn "Successfully recovered repository!"
 				putStrLn "Please carefully check that the changes mentioned above are ok.."
-				return (True, stillmissing, modifiedbranches)
+				return (True, modifiedbranches)
 	
 	corruptedindex = do
 		nukeFile (indexFile g)
 		-- The corrupted index can prevent fsck from finding other
 		-- problems, so re-run repair.
 		fsckresult' <- findBroken False g
-		result <- runRepairOf fsckresult' forced referencerepo g
+		result <- runRepairOf fsckresult' removablebranch forced referencerepo g
 		putStrLn "Removed the corrupted index file. You should look at what files are present in your working tree and git add them back to the index when appropriate."
 		return result
 
-	successfulfinish stillmissing modifiedbranches = do
+	successfulfinish modifiedbranches = do
 		mapM_ putStrLn
 			[ "Successfully recovered repository!"
 			, "You should run \"git fsck\" to make sure, but it looks like everything was recovered ok."
 			]
-		return (True, stillmissing, modifiedbranches)
-	unsuccessfulfinish stillmissing = do
+		return (True, modifiedbranches)
+	unsuccessfulfinish = do
 		if repoIsLocalBare g
 			then do
 				putStrLn "If you have a clone of this bare repository, you should add it as a remote of this repository, and retry."
 				putStrLn "If there are no clones of this repository, you can instead retry with the --force parameter to force recovery to a possibly usable state."
-				return (False, stillmissing, [])
-			else needforce stillmissing
-	needforce stillmissing = do
+				return (False, [])
+			else needforce
+	needforce = do
 		putStrLn "To force a recovery to a usable state, retry with the --force parameter."
-		return (False, stillmissing, [])
+		return (False, [])
 
-successfulRepair :: (Bool, MissingObjects, [Branch]) -> Bool
-successfulRepair = fst3
+successfulRepair :: (Bool, [Branch]) -> Bool
+successfulRepair = fst
 
 safeReadFile :: FilePath -> IO String
 safeReadFile f = do
